/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2008 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include "dcb_osdep.h"
#include "tlv.h"
#include "l2_packet.h"
#include "dcb_protocol.h"

int tlv_ok(struct unpacked_tlv *tlv)
{
	if (!tlv || (!tlv->length && tlv->type))
		return 0;
	else
		return 1;
}

struct packed_tlv *pack_tlv(struct unpacked_tlv *tlv)
{
	u16 tl = 0;
	struct packed_tlv *pkd_tlv = NULL;

	if (!tlv_ok(tlv))
		return NULL;

	tl = tlv->type;
	tl = tl << 9;
	tl |= tlv->length & 0x01ff;
	tl = htons(tl);

	pkd_tlv = (struct packed_tlv *)malloc(sizeof(struct packed_tlv));
	if(!pkd_tlv) {
		printf("pack_tlv: Failed to malloc pkd_tlv\n");
		return NULL;
	}
	memset(pkd_tlv,0,sizeof(struct packed_tlv));
	pkd_tlv->size = tlv->length + sizeof(tl);
	pkd_tlv->tlv = (u8 *)malloc(pkd_tlv->size);
	if(pkd_tlv->tlv) {
		memset(pkd_tlv->tlv,0, pkd_tlv->size);
		memcpy(pkd_tlv->tlv, &tl, sizeof(tl));
		if (tlv->length)
			memcpy(&pkd_tlv->tlv[sizeof(tl)], tlv->info_string,
				tlv->length);
	} else {
		printf("pack_tlv: Failed to malloc tlv\n");
		free(pkd_tlv);
		pkd_tlv = NULL;
		return NULL;
	}
	return pkd_tlv;
}


struct unpacked_tlv *unpack_tlv(struct packed_tlv *tlv)
{
	u16 tl = 0;
	struct unpacked_tlv *upkd_tlv = NULL;

	if(!tlv) {
		return NULL;
	}

	memcpy(&tl,tlv->tlv, sizeof(tl));
	tl = ntohs(tl);

	upkd_tlv = (struct unpacked_tlv *)malloc(sizeof(struct unpacked_tlv));
	if(upkd_tlv) {
		memset(upkd_tlv,0, sizeof(struct unpacked_tlv));
		upkd_tlv->length = tl & 0x01ff;
		upkd_tlv->info_string = (u8 *)malloc(upkd_tlv->length);
		if(upkd_tlv->info_string) {
			memset(upkd_tlv->info_string,0, upkd_tlv->length);
			tl = tl >> 9;
			upkd_tlv->type = (u8)tl;
			memcpy(upkd_tlv->info_string, &tlv->tlv[sizeof(tl)],
				upkd_tlv->length);
		} else {
			printf("unpack_tlv: Failed to malloc info_string\n");
			free (upkd_tlv);
			return NULL;
		}
	} else {
		printf("unpack_tlv: Failed to malloc upkd_tlv\n");
		return NULL;
	}
	return upkd_tlv;
}

struct unpacked_tlv *destruct_unpacked_tlv(struct unpacked_tlv *tlv)
{
	if (tlv != NULL) {
		if (tlv->info_string != NULL) {
			free(tlv->info_string);
			tlv->info_string = NULL;
		}
		free(tlv);
		tlv = NULL;
	}
	return NULL;
}

struct packed_tlv *destruct_packed_tlv(struct packed_tlv *tlv)
{
	if (tlv != NULL) {
		if (tlv->tlv != NULL) {
			free(tlv->tlv);
			tlv->tlv = NULL;
		}
		free(tlv);
		tlv = NULL;
	}
	return NULL;
}

struct unpacked_tlv *create_tlv()
{
	struct unpacked_tlv *tlv =
		(struct unpacked_tlv *)malloc(sizeof(struct unpacked_tlv));

	if(tlv) {
		memset(tlv,0, sizeof(struct unpacked_tlv));
		return tlv;
	} else {
		printf("create_tlv: Failed to malloc tlv\n");
		return NULL;
	}
}

struct unpacked_tlv *construct_type0_tlv()
{
	struct unpacked_tlv *tlv = create_tlv();

	if(tlv) {
		tlv->type = END_OF_LLDPDU_TLV;
		tlv->length = 0;
		tlv->info_string = NULL;
	}
	return tlv;
}

boolean_t is_validate_type0_tlv(struct unpacked_tlv *tlv)
{
	return 0;
}

struct unpacked_tlv *construct_type1_tlv(struct port *port)
{
	struct unpacked_tlv *tlv = create_tlv();

	if (!tlv)
		return NULL;

	tlv->type = CHASSIS_ID_TLV;
	tlv->length = (u16)(strlen(port->ifname)) + sizeof(u8);
	tlv->info_string = (u8 *)malloc(tlv->length);
	if(tlv->info_string) {
		memset(tlv->info_string,0, tlv->length);
		tlv->info_string[0] = CHASSIS_ID_INTERFACE_NAME;
		memcpy(tlv->info_string+1, port->ifname, tlv->length-1);
	} else {
		printf("construct_type1_tlv: Failed to malloc info_string\n");
		free(tlv);
		return NULL;
	}
	return tlv;
}

boolean_t is_validate_type1_tlv(struct unpacked_tlv *tlv)
{
	return 0;
}

struct unpacked_tlv *construct_type2_tlv(struct port *port)
{
	struct unpacked_tlv *tlv = create_tlv();
	u8 own_addr[ETH_ALEN], *hwaddr;

	if (!tlv)
		return NULL;

	hwaddr = &own_addr[0];
	tlv->type = PORT_ID_TLV;
	tlv->length = 7;
	tlv->info_string = (u8 *)malloc(tlv->length);
	if(tlv->info_string){
		memset(tlv->info_string,0, tlv->length);
		tlv->info_string[0] = PORT_ID_MAC_ADDRESS;
		if (l2_packet_get_own_addr(port->l2, hwaddr) >= 0)
			memcpy(&tlv->info_string[1], &own_addr, ETH_ALEN);
	} else {
		printf("construct_type2_tlv: Failed to malloc info_string\n");
		free(tlv);
		return NULL;
	}
	return tlv;
}

boolean_t is_validate_type2_tlv(struct unpacked_tlv *tlv)
{
	return 0;
}

struct unpacked_tlv *construct_type3_tlv(struct port *port)
{
	struct unpacked_tlv *tlv = create_tlv();

	if (!tlv)
		return NULL;

	tlv->type = TIME_TO_LIVE_TLV;
	tlv->length = 2;
	tlv->info_string = (u8 *)malloc(tlv->length);
	if(tlv->info_string) {
		memset(tlv->info_string,0, tlv->length);
		memcpy(tlv->info_string, &(port->porttx.txTTL), tlv->length);
	} else {
		printf("construct_type3_tlv: Failed to malloc info_string\n");
		free(tlv);
		return NULL;
	}
	return tlv;
}

boolean_t is_validate_type3_tlv(struct unpacked_tlv *tlv)
{
	return 0;
}

struct unpacked_tlv *construct_type127_tlv(struct port *port)
{
	struct unpacked_tlv *tlv = create_tlv();
	struct  packed_tlv *ptlv =  NULL;
	u8 dcb_subtype = 1;
	u8 oui[DCB_OUI_LEN] = {0x00,0x1b,0x21};
	u32 offset = 0;

	if (!tlv)
		return NULL;

	tlv->type = ORG_SPECIFIC_TLV;
	tlv->length = DCB_OUI_LEN + OUI_SUBTYPE_LEN;
	if (port->porttlvs.control) {
		tlv->length = tlv->length + port->porttlvs.control->length;
		if (port->porttlvs.control->length) tlv->length+=2;
	}
	if (port->porttlvs.pg) {
		tlv->length = tlv->length + port->porttlvs.pg->length;
		if (port->porttlvs.pg->length) tlv->length+=2;
	}
	if (port->porttlvs.pfc) {
		tlv->length = tlv->length + port->porttlvs.pfc->length;
		if (port->porttlvs.pfc->length) tlv->length+=2;
	}
	if (port->porttlvs.app) {
		tlv->length = tlv->length + port->porttlvs.app->length;
		tlv->length+=2;
	}

	tlv->info_string = (u8 *)malloc(tlv->length);
	if (tlv->info_string == NULL)
		goto error;
	memset(tlv->info_string,0, tlv->length);

	if ((DCB_OUI_LEN + OUI_SUBTYPE_LEN) > tlv->length)
		goto error;
	memcpy(tlv->info_string, &oui, DCB_OUI_LEN);
	offset += DCB_OUI_LEN;
	memcpy(&tlv->info_string[offset], &dcb_subtype, OUI_SUBTYPE_LEN);
	offset += OUI_SUBTYPE_LEN;

	if (tlv_ok(port->porttlvs.control)) {
		ptlv =  pack_tlv(port->porttlvs.control);
		if (!ptlv || ((ptlv->size+offset) > tlv->length))
			goto error;
		memcpy(&tlv->info_string[offset], ptlv->tlv, ptlv->size);
		offset += ptlv->size;
		ptlv = destruct_packed_tlv(ptlv);
	}
	if (tlv_ok(port->porttlvs.pg)) {
		ptlv = pack_tlv(port->porttlvs.pg);
		if (!ptlv || ((ptlv->size+offset) > tlv->length))
			goto error;
		memcpy(&tlv->info_string[offset], ptlv->tlv, ptlv->size);
		offset += ptlv->size;
		ptlv = destruct_packed_tlv(ptlv);
	}
	if (tlv_ok(port->porttlvs.pfc)) {
		ptlv = pack_tlv(port->porttlvs.pfc);
		if (!ptlv || ((ptlv->size+offset) > tlv->length))
			goto error;
		memcpy(&tlv->info_string[offset], ptlv->tlv, ptlv->size);
		offset += ptlv->size;
		ptlv = destruct_packed_tlv(ptlv);
	}
	if (tlv_ok(port->porttlvs.app)) {
		ptlv = pack_tlv(port->porttlvs.app);
		if (!ptlv || ((ptlv->size+offset) > tlv->length))
			goto error;
		memcpy(&tlv->info_string[offset], ptlv->tlv, ptlv->size);
		offset += ptlv->size;
		ptlv = destruct_packed_tlv(ptlv);
	}
	if (offset != tlv->length) printf("assert offset == tlv->length\n");
	assert(offset == tlv->length);
	return tlv;

error:
	if (tlv) {
		if (tlv->info_string)
			free(tlv->info_string);
		free(tlv);
	}
	printf("construct_type127_tlv: malloc failure \n");
	return NULL;
}

boolean_t is_validate_type127_tlv(struct unpacked_tlv *tlv)
{
	return 0;
}

struct unpacked_tlv *construct_control_tlv(struct port *port)
{
	struct unpacked_tlv *tlv = create_tlv();
	control_protocol_attribs control_setting;
	int result,i;
	u8 oper_version;
	u8 max_version;
	u32 seqno;
	u32 ackno;

	if (!tlv)
		return NULL;

	result = get_control(port->ifname, &control_setting);
	oper_version = (u8)control_setting.Oper_version;
	max_version = (u8)control_setting.Max_version;
	seqno = htonl(control_setting.SeqNo);
	ackno = htonl(control_setting.AckNo);

	tlv->type = DCB_CONTROL_TLV;
	tlv->length = sizeof(oper_version) + sizeof(max_version)
			+ sizeof(seqno) + sizeof(ackno);
	tlv->info_string = (u8 *)malloc(tlv->length);
	if (tlv->info_string) {
		memset(tlv->info_string,0, tlv->length);
		i = 0;
		memcpy(tlv->info_string,&oper_version, sizeof(oper_version));
		i += sizeof(oper_version);
		memcpy(tlv->info_string + i, &max_version, sizeof(max_version));
		i += sizeof(max_version);
		memcpy(tlv->info_string + i, &seqno, sizeof(seqno));
		i += sizeof(seqno);
		memcpy(tlv->info_string + i,&ackno ,sizeof(ackno));
		i = 0;
	} else {
		printf("construct_control_tlv: Failed to malloc info_string\n");
		free(tlv);
		return NULL;
	}

	return tlv;
}


struct unpacked_tlv *construct_pg_tlv(struct port *port, boolean_t *success)
{
	struct dcb_pg_info *pg_info;
	struct unpacked_tlv *tlv = create_tlv();
	pg_attribs  pg_setting;
	int result,i;

	*success = FALSE;
	if (!tlv) {
		return NULL;
	}
	result = get_pg(port->ifname, &pg_setting);
	if (result == dcb_success) {
		mark_pg_sent(port->ifname);
		if (!(pg_setting.protocol.Advertise)) {
			free(tlv);
			*success = TRUE;
			return NULL;
		}
	} else {
		free(tlv);
		return NULL;
	}

	pg_info = (struct dcb_pg_info *)malloc(sizeof(struct dcb_pg_info));
	if (pg_info) {
		memset(pg_info, 0, sizeof(struct dcb_pg_info));
		pg_info->hdr.oper_version =
			(u8)pg_setting.protocol.Oper_version;
		pg_info->hdr.max_version = (u8)pg_setting.protocol.Max_version;
		/* ewe Enable Willing Error */
		if (pg_setting.protocol.Enable == TRUE)
			pg_info->hdr.ewe |= BIT7;
		if (pg_setting.protocol.Willing == TRUE)
			pg_info->hdr.ewe |= BIT6;
		if (pg_setting.protocol.Error == TRUE)
			pg_info->hdr.ewe |= BIT5;
		pg_info->hdr.sub_type = DEFAULT_SUBTYPE;
		for (i = 0; i < 8; i++) {
			pg_info->desired_result.bwg_percentages[i] =
				pg_setting.tx.bwg_percent_of_link[i];
			u8 tmpbyte = 0;
			tmpbyte  = pg_setting.tx.up[i].bandwidth_group_mapping;
			tmpbyte = tmpbyte << 5;
			u8 tmpprio = 0;
			tmpprio = (u8)pg_setting.tx.up[i].strict_priority;
			tmpprio = tmpprio << 3;
			tmpbyte |= tmpprio;
			pg_info->desired_result.pg_up_settings[i].byte1 =
				tmpbyte;
			pg_info->desired_result.pg_up_settings[i].byte2 =
				pg_setting.tx.up[i].percent_of_bandwidth_group_cap;
		}
		tlv->length = sizeof(struct dcb_pg_info);
	} else {
		printf("construct_pg_tlv: Failed to malloc pg_info\n");
		free(tlv);
		return NULL;
	}
	tlv->type = DCB_PRIORITY_GROUPS_TLV;
	tlv->info_string = (u8 *)pg_info;
	*success = TRUE;
	return tlv;
}

struct unpacked_tlv *construct_pfc_tlv(struct port *port, boolean_t *success)
{
	struct dcb_pfc_info *pfc_info;
	struct unpacked_tlv *tlv = create_tlv();
	pfc_attribs pfc_setting;
	int result,i;

	*success = FALSE;
	if (!tlv)
		return NULL;
	result = get_pfc(port->ifname, &pfc_setting);
	if (result == dcb_success) {
		mark_pfc_sent(port->ifname);
		if (!(pfc_setting.protocol.Advertise)) {
			free(tlv);
			*success = TRUE;
			return NULL;
		}
	} else {
		free(tlv);
		return NULL;
	}

	pfc_info = (struct dcb_pfc_info *)malloc(sizeof(struct dcb_pfc_info));
	if (pfc_info) {
		memset(pfc_info,0,sizeof(struct  dcb_pfc_info));
		pfc_info->hdr.oper_version =
			(u8)pfc_setting.protocol.Oper_version;
		pfc_info->hdr.max_version
			= (u8)pfc_setting.protocol.Max_version;
		/* ewe Enable Willing Error */
		if(pfc_setting.protocol.Enable == TRUE)
			pfc_info->hdr.ewe |= BIT7;
		if(pfc_setting.protocol.Willing == TRUE)
			pfc_info->hdr.ewe |= BIT6;
		if(pfc_setting.protocol.Error == TRUE)
			pfc_info->hdr.ewe |= BIT5;
		pfc_info->hdr.sub_type = DEFAULT_SUBTYPE;
		u8 temp = 0;
		for(i = 0; i < 8; i++) {
			temp = (u8)(pfc_setting.admin[i] << i);
			pfc_info->desired_result.admin_map |= temp;
		}
		tlv->length = sizeof(struct dcb_pfc_info);
	} else {
		printf("construct_pfc_tlv: Failed to malloc pfc_info\n");
		free(tlv);
		return NULL;
	}
	tlv->type = DCB_PRIORITY_FLOW_CONTROL_TLV;
	tlv->info_string = (u8 *)pfc_info;
	*success = TRUE;
	return tlv;
}

struct unpacked_tlv *construct_app_tlv(struct port *port,
					u32 sub_type,
					boolean_t *success)
{
	struct dcb_app_info *app_info;
	struct unpacked_tlv *tlv = create_tlv();
	app_attribs     app_setting;
	int result;
	u32 i,len;

	*success = FALSE;
	if (!tlv)
		return NULL;

	memset(&app_setting, 0, sizeof(app_setting));
	result = get_app(port->ifname, sub_type, &app_setting);
	if (result == dcb_success) {
		mark_app_sent(port->ifname, sub_type);
		if (!(app_setting.protocol.Advertise)) {
			free(tlv);
			*success = TRUE;
			return NULL;
		}
	} else {
		free(tlv);
		return NULL;
	}
	len = sizeof(struct  dcb_tlv_header) + app_setting.Length;
	app_info = (struct dcb_app_info *)malloc(len);
	if (app_info) {
		memset(app_info,0,sizeof(struct  dcb_app_info));
		app_info->hdr.oper_version =
			(u8)app_setting.protocol.Oper_version;
		app_info->hdr.max_version =
			(u8)app_setting.protocol.Max_version;
		/* ewe Enable Willing Error */
		if(app_setting.protocol.Enable == TRUE)
			app_info->hdr.ewe |= BIT7;
		if(app_setting.protocol.Willing == TRUE)
			app_info->hdr.ewe |= BIT6;
		if(app_setting.protocol.Error == TRUE)
			app_info->hdr.ewe |= BIT5;
		app_info->hdr.sub_type = (u8)sub_type;
		for (i = 0; i < (int)app_setting.Length; i++) {
			app_info->desired_result[i] = app_setting.AppData[i];
		}
		tlv->length = (u16)len;
	} else {
		printf("construct_app_tlv: Failed to malloc app_info\n");
		free(tlv);
		return NULL;
	}
	tlv->type = DCB_APPLICATION_TLV;
	tlv->info_string = (u8 *)app_info;
	*success = TRUE;
	return tlv;
}
